import sys
from collections.abc import Collection, Iterable, Iterator
from itertools import chain
from typing import Any, Generic, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from csnlp import Solution, wrappers
from csnlp.core.cache import invalidate_caches_of
from csnlp.util.io import SupportsDeepcopyAndPickle
from csnlp.wrappers import Mpc
from gymnasium import Env
from gymnasium.spaces import Box

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from ...core.callbacks import AgentCallbackMixin
from ...core.exploration import ExplorationStrategy, NoExploration
from ...core.warmstart import WarmStartStrategy
from ...util.named import Named
from ...util.seeding import RngType, mk_seed

SymType = TypeVar("SymType", cs.SX, cs.MX)
ActType: TypeAlias = Union[npt.ArrayLike, dict[str, npt.ArrayLike]]
ObsType: TypeAlias = Union[npt.ArrayLike, dict[str, npt.ArrayLike]]


def _update_dicts(sinks: Iterable[dict], source: dict) -> Iterator[dict]:
    """Internal utility for updating dicts `sinks` with one `source`."""
    for sink in sinks:
        sink.update(source)
        yield sink


class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbackMixin, Generic[SymType]):
    r"""Simple MPC-based agent with a fixed (i.e., non-learnable) MPC controller.

    In this agent, the MPC controller parametrized in :math:`\theta` is used as policy
    provider, as well as to provide the value function :math:`V_\theta(s)` and quality
    function :math:`Q_\theta(s,a)`, where :math:`s` and :math:`a` are the state of the
    environment and a generic action, respectively. Since it only supports a fixed
    parametrization, this class does not use any RL or other learning method to improve
    its MPC policy.

    Parameters
    ----------
    mpc : :class:`csnlp.wrappers.Mpc`
        The MPC controller used as policy provider by this agent. The instance is
        modified in place to create the approximations of the state function
        :math:`V_\theta(s)` and action value function :math:`Q_\theta(s,a)`, so it is
        recommended not to modify it further after initialization of the agent.
        Moreover, some parameter and constraint names will need to be created, so an
        error is thrown if these names are already in use in the mpc.
    fixed_parameters : dict of (str, array_like) or collection of, optional
        A dict (or collection of dict, in case of the ``mpc`` wrapping an underlying
        :class:`csnlp.multistart.MultistartNlp` instance) whose keys are the names of
        the MPC parameters and the values are their corresponding values. Use this to
        specify fixed parameters, that is, non-learnable. If ``None``, then no fixed
        parameter is assumed.
    exploration : :class:`core.exploration.ExplorationStrategy`, optional
        Exploration strategy for inducing exploration in the online MPC policy. By
        default ``None``, in which case :class:`core.exploration.NoExploration` is used.
    warmstart : "last" or "last-successful" or WarmStartStrategy, optional
        The warmstart strategy for the MPC's NLP. If ``"last-successful"``, the last
        successful solution is used to warmstart the solver for the next iteration. If
        ``"last"``, the last solution is used, regardless of success or failure.
        Furthermore, an instance of :class:`core.warmstart.WarmStartStrategy` can
        be passed to specify a strategy for generating multiple warmstart points for the
        MPC's NLP instance. This is useful to generate multiple initial conditions for
        highly non-convex, nonlinear problems. This feature can only be used with an
        MPC that has an underlying multistart NLP problem (see :mod:`csnlp.multistart`).
    use_last_action_on_fail : bool, optional
        In case the MPC solver fails
         - if ``False``, the action from the last solver's iteration is returned anyway
           (though suboptimal)
         - if ``True``, the action from the last successful call to the MPC is returned
           instead (if the MPC has been solved at least once successfully).

        By default, ``False``.
    remove_bounds_on_initial_action : bool, optional
        When ``True``, the upper and lower bounds on the initial action are removed in
        the action-value function approximator :math:`Q_\theta(s,a)` since the first
        action is constrained to be equal to the provided action :math:`a`. This is
        useful to avoid issues in the LICQ of the NLP. However, it can lead to numerical
        problems. By default, ``False``.
    name : str, optional
        Name of the agent. If ``None``, one is automatically created from a counter of
        the class' instancies.

    Raises
    ------
    ValueError
        Raises if
         - the given ``mpc`` has no control action as optimization variable
         - the reserved parameter and constraint names are already in use (see
           :attr:`cost_perturbation_parameter`, :attr:`init_action_parameter` and
           :attr:`init_action_constraint`)
         - a multistart ``mpc`` is given, but the warmstart strategy ``warmstart`` asks
           for an incompatible number of starting points to be generated
         - a warmstart strategy ``warmstart`` is given, but the ``mpc`` does not have an
           underlying multistart NLP problem, so it cannot handle multiple starting
           points (see :attr:`csnlp.Nlp.is_multi` and
           :attr:`csnlp.multistart.MultistartNlp.is_multi`).
    """

    cost_perturbation_method = "normal"
    r"""The name of the method from :class`numpy.random.Generator` to be used to
    generate perturbations of the cost function in the state value function
    :math:`V_\theta(s)`."""

    cost_perturbation_parameter = "cost_perturbation"
    r"""The name of the parameter to be added to the original ``mpc`` problem for
    perturbing the state value function :math:`V_\theta(s)`."""

    init_action_parameter = "a_init"
    r"""Name of the parameter to be added to the original ``mpc`` problem for
    constraining the first action to be equal to :math:`a` in the action value function
    :math:`Q_\theta(s,a)`."""

    init_action_constraint = init_action_parameter
    r"""Name of the equality constraint to be added to the original ``mpc`` problem for
    constraining the first action to be equal to :math:`a` in the action value function
    :math:`Q_\theta(s,a)`."""

    def __init__(
        self,
        mpc: Mpc[SymType],
        fixed_parameters: Union[
            None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]
        ] = None,
        exploration: Optional[ExplorationStrategy] = None,
        warmstart: Union[
            Literal["last", "last-successful"], WarmStartStrategy
        ] = "last-successful",
        use_last_action_on_fail: bool = False,
        remove_bounds_on_initial_action: bool = False,
        name: Optional[str] = None,
    ) -> None:
        if isinstance(warmstart, str):
            warmstart = WarmStartStrategy(warmstart)
        ws_points = warmstart.n_points
        if mpc.is_multi and ws_points != 0 and mpc.nlp.starts - ws_points not in (0, 1):
            raise ValueError(
                "A multistart MPC was given with {mpc.nlp.starts} multistarts, but the "
                f"given warmstart strategy asks for {ws_points} starting points. "
                "Expected either 0 warmstart points (i.e., it is disabled), or the same"
                " number as MPC's multistarts, or at most one less."
            )
        elif not mpc.is_multi and ws_points > 0:
            raise ValueError(
                "Got a warmstart strategy with more than 0 starting points, but the "
                "given does not have an underlying multistart NLP problem."
            )
        Named.__init__(self, name)
        SupportsDeepcopyAndPickle.__init__(self)
        AgentCallbackMixin.__init__(self)
        self._fixed_pars = fixed_parameters
        if exploration is None:
            exploration = NoExploration()
        self._exploration = exploration
        self._warmstart = warmstart
        self._last_action_on_fail = use_last_action_on_fail
        self._last_solution: Optional[Solution[SymType]] = None
        self._last_action: Optional[cs.DM] = None
        self._V, self._Q = self._setup_V_and_Q(mpc, remove_bounds_on_initial_action)
        self._post_setup_V_and_Q()

    @property
    def unwrapped(self) -> "Agent":
        """Gets the underlying wrapped instance of an agent. In this case, since the
        agent is not wrapped at all, returns itself."""
        return self

    def is_wrapped(self, *args: Any, **kwargs: Any) -> bool:
        """Gets whether the agent instance is wrapped or not by the wrapper type.

        Returns
        -------
        bool
            A flag indicating whether the agent is wrapped or not.
        """
        return False

    @property
    def V(self) -> Mpc[SymType]:
        r"""Gets the MPC function approximation of the state value function
        :math:`V_\theta(s)`."""
        return self._V

    @property
    def Q(self) -> Mpc[SymType]:
        r"""Gets the MPC function approximation of the action value function
        :math:`Q_\theta(s,a)`."""
        return self._Q

    @property
    def fixed_parameters(
        self,
    ) -> Union[None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]]:
        """Gets the fixed parameters of the MPC controller, i.e., the non-learnable
        ones.

        Returns
        -------
        ``None`` or dict of (str, array_like), or collection of
            The returned object can be either
             - ``None``, if the MPC controller has no fixed parameters
             - a dict whose keys are the names of the MPC parameters and the values are
               their corresponding values, when the MPC controller wraps an instance of
               :class:`csnlp.Nlp`, or it wraps an instance of
               :class:`csnlp.multistart.MultistartNlp` but the same set of parameters
               is meant to be used for all scenarios
             - a collection of such dictionaries, when the MPC controller wraps an
               instance of :class:`csnlp.multistart.MultistartNlp` and different
               parameters are meant to be used for each scenario.
        """
        return self._fixed_pars

    @property
    def exploration(self) -> ExplorationStrategy:
        r"""Gets the exploration strategy used within this agent to perturb the
        policy provided by the MPC controller via :math:`V_\theta(s)`."""
        return self._exploration

    @property
    def warmstart(self) -> WarmStartStrategy:
        """Gets the warmstart strategy used within this agent. This strategy is used to
        generate the initial guess for the solver to optimize the MPC's NLP."""
        return self._warmstart

    def reset(self, seed: RngType = None) -> None:
        """Resets the agent. This includes resetting the warmstart strategy, the
        exploration strategy, and the some internal variables of the agent.

        Parameters
        ----------
        seed : RngType, optional
            The seed to reset the  :class:`numpy.random.Generator` instances. By
            default, ``None``.
        """
        self._last_solution = None
        self._last_action = None
        self.warmstart.reset(seed)
        self.exploration.reset(seed)

    def _solve_mpc(
        self,
        mpc: Mpc[SymType],
        state: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        action: Union[None, npt.ArrayLike, dict[str, npt.ArrayLike]] = None,
        perturbation: Optional[npt.ArrayLike] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        store_solution: bool = True,
    ) -> Solution[SymType]:
        r"""Solves the agent's specific MPC optimal control problem.

        Parameters
        ----------
        mpc : Mpc
            The MPC problem to solve, either :attr:`V` or :attr:`Q`.
        state : array_like or dict of (str, array_like)
            The initial state at which to evaluate the MPC policy, i.e., :math:`s` in
            :math:`V_\theta(s)` or :math:`Q_\theta(s,a)`. It can be either a 1D array
            representing the value of all initial states of the MPC, concatenated.
            Otherwise, a dict whose keys are the names of each state, and values are
            their numerical initial state values.
        action : array_like or dict of (str, array_like), optional
            Same for ``state``, but for the action, i.e., the initial action at which to
            evaluate the MPC action value function, i.e., :math:`a` in
            :math:`Q_\theta(s,a)`. Obviously, it is only pertinent if ``mpc`` is
            :attr:`Q`, while it should be ``None`` for :attr:`V`.
        perturbation : array_like, optional
            The **gradient-based** cost perturbation used to induce exploration in
            :math:`V_\theta(s)`. Should be ``None`` for :math:`Q_\theta(s,a)`, or in
            case of other types of exploration are used.
        vals0 : dict of (str, array_like) or iterable of, optional
            A dict (or an iterable of dict, in case of
            :class:`csnlp.multistart.MultistartNlp` is used), whose
            keys are the names of the MPC variables, and values are the numerical
            initial values of each variable. Use this argument to warmstart the MPC.
            If ``None``, and a previous solution (possibly, successful) is available,
            the MPC solver is automatically warmstarted. If an iterable is passed
            instead, the warmstarting strategy is bypassed.
        store_solution : bool, optional
            By default, the MPC solution is stored accordingly to the :attr:`warmstart`
            strategy. If set to ``False``, this flag allows to disable the behaviour for
            this particular solution.

        Returns
        -------
        Solution
            The solution of the MPC.
        """
        # convert state keys into initial state keys (i.e., with "_0")
        if isinstance(state, dict):
            x0_dict = {f"{k}_0": v for k, v in state.items()}
        else:
            mpcstates = mpc.initial_states
            if len(mpcstates) == 1:
                states = (state,)
            else:
                cumsizes = np.cumsum([s.shape[0] for s in mpcstates.values()][:-1])
                states = np.split(state, cumsizes)
            x0_dict = dict(zip(mpcstates.keys(), states))

        # convert action dict to vector if not None
        if action is None:
            u0_vec = None
        elif isinstance(action, dict):
            u0_vec = cs.vertcat(*(action[k] for k in mpc.actions.keys()))
        else:
            u0_vec = action

        # merge (initial) state, action and perturbation in unique dict
        additional_pars = x0_dict
        if u0_vec is not None:
            additional_pars[self.init_action_parameter] = u0_vec
        if self.cost_perturbation_parameter in mpc.parameters:
            additional_pars[self.cost_perturbation_parameter] = (
                0 if perturbation is None else perturbation
            )

        # create pars and vals0
        pars = self._get_parameters()
        if pars is None:
            pars = additional_pars
        elif isinstance(pars, dict):
            pars.update(additional_pars)
        else:  # iterable of dict
            pars = _update_dicts(pars, additional_pars)
        if vals0 is None and self._last_solution is not None:
            vals0 = self._last_solution.vals

        # use the warmstart strategy to generate multiple initial points for the NLP if
        # the NLP supports multi and `vals0` is not already an iterable of dict
        if mpc.is_multi and (vals0 is None or isinstance(vals0, dict)):
            more_vals0s = self._warmstart.generate(vals0)
            if mpc.nlp.starts > self._warmstart.n_points:
                # the difference between these two has been checked to be at most one,
                # meaning we can include `vals0` itself
                more_vals0s = chain((vals0,), more_vals0s)
            vals0 = more_vals0s

        # solve and store solution
        sol = mpc(pars, vals0)
        if store_solution and (self._warmstart.store_always or sol.success):
            self._last_solution = sol
        return sol

    def state_value(
        self,
        state: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        deterministic: bool = False,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        action_space: Optional[Box] = None,
        **kwargs,
    ) -> tuple[cs.DM, Solution[SymType]]:
        r"""Computes the MPC-based state value function approximation
        :math:`V_\theta(s)`.

        Parameters
        ----------
        state : array_like or dict of (str, array_like)
            The initial state at which to evaluate the MPC approximation of the state
            value function, i.e., :math:`s` in :math:`V_\theta(s)`. It can be either a
            1D array representing the value of all initial states of the MPC,
            concatenated. Otherwise, a dict whose keys are the names of each state, and
            values are their numerical initial state values.
        deterministic : bool, optional
            If ``False``, the MPC controller is perturbed according to the
            :attr:`exploration` strategy to induce some exploratory behaviour.
            Otherwise, no perturbation is performed. By default, ``False``.
        vals0 : dict of (str, array_like) or iterable of, optional
            A dict (or an iterable of dict, in case of
            :class:`csnlp.multistart.MultistartNlp` is used), whose
            keys are the names of the MPC variables, and values are the numerical
            initial values of each variable. Use this argument to warmstart the MPC.
            If ``None``, and a previous solution (possibly, successful) is available,
            the MPC solver is automatically warmstarted. If an iterable is passed
            instead, the warmstarting strategy is bypassed.
        action_space : gymnasium.spaces.Box, optional
            The action space of the environment the agent is being evaluated/trained on.
            If not ``None``, it is used in case an additive exploration perturbation is
            summed to the action in order to clip it back into the action space.

        Returns
        -------
        casadi.DM
            The first optimal action according to the solution of the state value
            function, possibly perturbed by exploration noise, i.e.,

            .. math:: u_0^\star = \arg\min_{u} V_\theta(s)

        Solution
            The solution of the MPC approximation :math:`V_\theta(s)` at the given
            state.
        """
        V = self._V
        exploration = self._exploration
        exploration_mode = exploration.mode
        na = V.na
        if deterministic or exploration_mode is None or not exploration.can_explore():
            pert = None
        else:
            pert = exploration.perturbation(self.cost_perturbation_method, size=(na, 1))

        grad_pert = pert if exploration_mode == "gradient-based" else None
        sol = self._solve_mpc(V, state, perturbation=grad_pert, vals0=vals0, **kwargs)
        action_opt = cs.vertcat(*(sol.vals[u][:, 0] for u in V.actions.keys()))

        if sol.success:
            self._last_action = action_opt
        elif self._last_action_on_fail and self._last_action is not None:
            action_opt = self._last_action

        if pert is not None and exploration_mode == "additive":
            action_opt_noisy = action_opt + pert
            if action_space is not None:
                lb = action_space.low.reshape(na, 1)
                ub = action_space.high.reshape(na, 1)
                action_opt_noisy = np.clip(action_opt_noisy, lb, ub)
                # if np.equal(action_opt_noisy, action_opt).all():
                #     action_opt_noisy = np.clip(action_opt - pert, lb, ub)
            action_opt = action_opt_noisy
        return action_opt, sol

    def action_value(
        self,
        state: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        action: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        **kwargs,
    ) -> Solution[SymType]:
        r"""Computes the MPC-based action value function approximation
        :math:`Q_\theta(s,a)`.

        Parameters
        ----------
        state : array_like or dict of (str, array_like)
            The initial state at which to evaluate the action value function, i.e.,
            :math:`s` in :math:`Q_\theta(s,a)`. It can be either a 1D array representing
            the value of all initial states of the MPC, concatenated. Otherwise, a dict
            whose keys are the names of each state, and values are their numerical
            initial state values.
        action : array_like or dict of (str, array_like), optional
            Same for ``state``, but for the action, i.e., the initial action at which to
            evaluate the MPC action value function, i.e., :math:`a` in
            :math:`Q_\theta(s,a)`.
        vals0 : dict of (str, array_like) or iterable of, optional
            A dict (or an iterable of dict, in case of
            :class:`csnlp.multistart.MultistartNlp` is used), whose
            keys are the names of the MPC variables, and values are the numerical
            initial values of each variable. Use this argument to warmstart the MPC.
            If ``None``, and a previous solution (possibly, successful) is available,
            the MPC solver is automatically warmstarted. If an iterable is passed
            instead, the warmstarting strategy is bypassed.

        Returns
        -------
        Solution
            The solution of the MPC approximation :math:`Q_\theta(s,a)` at the given
            state and action pair.
        """
        return self._solve_mpc(self._Q, state, action, vals0=vals0, **kwargs)

    def evaluate(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        deterministic: bool = True,
        seed: RngType = None,
        raises: bool = True,
        env_reset_options: Optional[dict[str, Any]] = None,
    ) -> npt.NDArray[np.floating]:
        r"""Evaluates the agent in a given environment.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            The gym environment where to evaluate the agent in.
        episodes : int
            Number of evaluation episodes.
        deterministic : bool, optional
            Whether the agent should act deterministically, i.e., applying no
            exploration to the policy provided by the MPC. By default, ``True``.
        seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
            Seed for the agent's and env's random number generator. By default ``None``.
        raises : bool, optional
            If ``True``, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.
        env_reset_options : dict, optional
            Additional information to specify how the environment is reset at each
            evalution episode (optional, depending on the specific environment).

        Returns
        -------
        array of doubles
            The cumulative returns (one return per evaluation episode).

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises if the MPC optimization solver fails and ``raises=True``.

        Notes
        -----
        After solving :math:`V_\theta(s)` for the current env's state `s`, the action
        is passed to the environment as the concatenation of the first optimal action
        variables of the MPC (see `csnlp.Mpc.actions`).
        """
        rng = np.random.default_rng(seed)
        self.reset(rng)
        returns = np.zeros(episodes)
        self.on_validation_start(env)

        for episode in range(episodes):
            state, _ = env.reset(seed=mk_seed(rng), options=env_reset_options)
            truncated, terminated, timestep = False, False, 0
            self.on_episode_start(env, episode, state)

            while not (truncated or terminated):
                action, sol = self.state_value(state, deterministic)
                if not sol.success:
                    self.on_mpc_failure(episode, timestep, sol.status, raises)

                state, r, truncated, terminated, _ = env.step(action)
                self.on_env_step(env, episode, timestep)

                returns[episode] += r
                timestep += 1
                self.on_timestep_end(env, episode, timestep)

            self.on_episode_end(env, episode, returns[episode])

        self.on_validation_end(env, returns)
        return returns

    def _setup_V_and_Q(
        self, mpc: Mpc[SymType], remove_bounds_on_initial_action: bool
    ) -> tuple[Mpc[SymType], Mpc[SymType]]:
        """Internal utility to setup the function approximators for the value function
        ``V(s)`` and the quality function ``Q(s,a)``."""
        na = mpc.na
        if na <= 0:
            raise ValueError(f"Expected Mpc with na>0; got na={na} instead.")

        # create V and Q function approximations
        V, Q = mpc, mpc.copy()
        V.unwrapped.name += "_V"
        Q.unwrapped.name += "_Q"

        # for Q, add the additional constraint on the initial action to be equal to a0,
        # and remove the now useless upper/lower bounds on the initial action
        a0 = Q.nlp.parameter(self.init_action_parameter, (na, 1))
        u0 = cs.vcat(mpc.first_actions.values())
        Q.nlp.constraint(self.init_action_constraint, u0, "==", a0)
        if remove_bounds_on_initial_action:
            for name, a in mpc.first_actions.items():
                na_ = a.size1()
                Q.nlp.remove_variable_bounds(name, "both", ((r, 0) for r in range(na_)))

        # for V, add the cost perturbation parameter (only if gradient-based)
        if self._exploration.mode == "gradient-based":
            perturbation = V.nlp.parameter(self.cost_perturbation_parameter, (na, 1))
            f = V.nlp.f
            if mpc.is_wrapped(wrappers.NlpScaling):
                f = mpc.scale(f)
            V.nlp.minimize(f + cs.dot(perturbation, u0))

        # invalidate caches for V and Q since some modifications have been done
        for nlp in (V, Q):
            nlp_ = nlp
            while nlp_ is not nlp_.unwrapped:
                invalidate_caches_of(nlp_)
                nlp_ = nlp_.nlp
            invalidate_caches_of(nlp_.unwrapped)
        return V, Q

    def _post_setup_V_and_Q(self) -> None:
        """Internal utility that is run after the creation of ``V`` and ``Q``, allowing
        for further customization in inheriting classes."""

    def _get_parameters(
        self,
    ) -> Union[None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]]:
        """Internal utility to retrieve parameters of the MPC in order to solve it.
        :class:`Agent` has no learnable parameter, so only fixed parameters are
        returned."""
        return self._fixed_pars

    def __deepcopy__(self, memo: Optional[dict[int, list[Any]]] = None) -> "Agent":
        """Ensures that the copy has a new name."""
        y = super().__deepcopy__(memo)
        if hasattr(y, "name"):
            y.name += "_copy"
        return y



"""A submodule providing both base classes for schedulers and some concrete
implementations for the most common cases, such as linearly and exponentially decaying
schedulers."""

import abc
from collections.abc import Iterable as _Iterable
from itertools import accumulate as _acc
from itertools import repeat as _rep
from operator import mul as _mul
from typing import Generic as _Generic
from typing import TypeVar as _TypeVar
from typing import Union

import numpy as np
import numpy.typing as npt

ScType = _TypeVar("ScType", int, float, npt.NDArray[np.int_], npt.NDArray[np.floating])


class Scheduler(abc.ABC, _Generic[ScType]):
    """Base abstract class for schedulers.

    Schedulers are helpful classes to update or decay different quantities, such as
    learning rates or exploration probability.

    Parameters
    ----------
    init_value : supports-algebraic-operations
        Initial value that will be updated by this scheduler.

    Notes
    -----
    If the scheduler has a final iteration, it is expected to raise a
    :class:`StopIteration` when the last iteration is reached.
    """

    def __init__(self, init_value: ScType) -> None:
        super().__init__()
        self.value = init_value

    @abc.abstractmethod
    def step(self) -> None:
        """Updates the value of the scheduler by one step.

        Raises
        ------
        StopIteration
            Raises if the final iteration of the scheduler (if any) has been reached and
            :meth:`step` was called again.
        """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.value})"

    def __str__(self) -> str:
        return self.__repr__()


class NoScheduling(Scheduler[ScType]):
    """Scheduler that actually performs no scheduling and holds the initial value
    indefinitely constant."""

    def step(self) -> None:
        return


class ExponentialScheduler(Scheduler[ScType]):
    r"""Exponentially decay scheduler.

    It exponentially decays the value of the scheduler by ``factor`` every step, i.e.,
    mathematically, at the :math:`k`-th step, the value :math:`v_k` is

    .. math:: v_k = v_0 f^k,

    where :math:`v_0` is the initial value and :math:`f` is the ``factor``.

    Parameters
    ----------
    init_value : supports-algebraic-operations
        Initial value that will be updated by this scheduler.
    factor : supports-algebraic-operations
        The exponential factor to decay the initial value with.
    """

    def __init__(self, init_value: ScType, factor: ScType) -> None:
        super().__init__(init_value)
        self.factor = factor

    def step(self) -> None:
        self.value *= self.factor

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.value},factor={self.factor})"


class LinearScheduler(Scheduler[ScType]):
    r"""Linear scheduling from initial to final value in a fixed number of steps.

    This scheduler linearly updates the initial value towards a final one to be reached
    in :math:`N` total steps, i.e., mathematically, at the :math:`k`-th step, the value
    :math:`v_k` is

    .. math:: v_k = v_0 + \left(v_N - v_0\right) \frac{k}{N},

    where :math:`v_0` is the initial value and :math:`v_0` the final.

    Parameters
    ----------
    init_value : supports-algebraic-operations
        Initial value that will be updated by this scheduler.
    final_value : supports-algebraic-operations
        Final value that will be reached by the scheduler after ``total_steps``.
    total_steps : int
        Total number of steps to linearly interpolate between ``init_value`` and
        ``final_value``.
    """

    def __init__(
        self, init_value: ScType, final_value: ScType, total_steps: int
    ) -> None:
        super().__init__(init_value)
        increment = (final_value - init_value) / total_steps
        self.init_value = init_value
        self.final_value = final_value
        self.total_steps = total_steps
        self.generator = _acc(_rep(increment, total_steps), initial=init_value)
        next(self.generator)

    def step(self) -> None:
        self.value = next(self.generator)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.value},x0={self.init_value},"
            f"xf={self.final_value},N={self.total_steps})"
        )


class LogLinearScheduler(ExponentialScheduler[ScType]):
    r"""Scheduler that updates the scheduled quantity from the initial to the final
    value in a logarithmic fashion.

    In other words, this scheduler updates the initial value of the scheduler towards
    the final one (to be reached in :math:`N` total steps) in a linear fashion between
    the exponents of the two. Mathematically, at the :math:`k`-th step, the value
    :math:`v_k` is

    .. math:: v_k = v_0 \exp\left(
            \text{ln} \left( \frac{v_N}{v_0} \right) \frac{k}{N}
        \right),

    where :math:`v_0` is the initial value, and :math:`v_N` the final.

    Parameters
    ----------
    init_value : supports-algebraic-operations
        Initial value that will be updated by this scheduler.
    final_value : supports-algebraic-operations
        Final value that will be reached by the scheduler after ``total_steps``.
    total_steps : int
        Total number of steps to log-linearly interpolate between ``init_value`` and
        ``final_value``.
    """

    def __init__(
        self, init_value: ScType, final_value: ScType, total_steps: int
    ) -> None:
        factor = np.exp(np.log(final_value / init_value) / total_steps)
        super().__init__(init_value, factor)
        self.init_value = init_value
        self.final_value = final_value
        self.total_steps = total_steps
        self.generator = _acc(_rep(factor, total_steps), _mul, initial=init_value)
        next(self.generator)

    def step(self) -> None:
        self.value = next(self.generator)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.value},x0={self.init_value},"
            f"xf={self.final_value},N={self.total_steps})"
        )


class Chain(Scheduler[ScType]):
    """Chains multiple schedulers together.

    Parameters
    ----------
    schedulers : iterable of schedulers or of (scheduler, int)
        An iterable of schedulers to be chained together. If an iterable of tuples is
        passed instead, the first element is expected to be an instance of scheduler,
        and the second one an integer indicating the number of steps to run that
        scheduler for. This is useful in case the scheduler has no fixed number of
        steps, such as :class:`ExponentialScheduler`.
    """

    def __init__(
        self,
        schedulers: _Iterable[Union[Scheduler[ScType], tuple[Scheduler[ScType], int]]],
    ) -> None:
        self.schedulers = iter(schedulers)
        self._next_scheduler()
        super().__init__(self._current_scheduler.value)

    def step(self) -> None:
        if self._current_steps is not None:
            self._current_scheduler.step()
            self._current_steps -= 1
            if self._current_steps == 0:
                self._next_scheduler()
        else:
            try:
                self._current_scheduler.step()
            except StopIteration:
                self._next_scheduler()
        self.value = self._current_scheduler.value

    def _next_scheduler(self) -> None:
        """Fetches the next scheduler in the chain. If the chain is over, raises
        :class:`StopIteration`, which is in line with the behaviour of other
        schedulers."""
        scheduler = next(self.schedulers)
        if isinstance(scheduler, tuple):
            self._current_scheduler, self._current_steps = scheduler
        else:
            self._current_scheduler = scheduler
            self._current_steps = None

    def __repr__(self) -> str:
        s = self._current_scheduler
        return f"{self.__class__.__name__}(x={self.value},current={s})"



import sys
from collections.abc import Collection, Iterable
from typing import Callable, Generic, Literal, Optional, SupportsFloat, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from csnlp import Solution
from csnlp.wrappers import Mpc, NlpSensitivity
from gymnasium import Env

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from ..core.experience import ExperienceReplay
from ..core.exploration import ExplorationStrategy
from ..core.parameters import LearnableParametersDict
from ..core.update import UpdateStrategy
from ..core.warmstart import WarmStartStrategy
from ..optim.gradient_based_optimizer import GradientBasedOptimizer
from .common.agent import ActType, ObsType, SymType
from .common.rl_learning_agent import LrType, RlLearningAgent

# the experience buffer contains the gradient and, possibly, the hessian of the Bellman
# residuals w.r.t. the learnable parameters theta
ExpType: TypeAlias = Union[
    npt.NDArray[np.floating], tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]
]


class LstdQLearningAgent(
    RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, LrType]
):
    r"""Second-order Least-Squares Temporal Difference (LSTD) Q-learning agent, as first
    proposed in a simpler format in :cite:`gros_datadriven_2020`, and then in
    :cite:`esfahani_approximate_2021`.

    The Q-learning agent uses an MPC controller as policy provider and function
    approximation, and adjusts its parametrization according to the temporal-difference
    error, with the goal of improving the policy, in an indirect fashion by learning the
    action value function.

    Parameters
    ----------
    mpc : :class:`csnlp.wrappers.Mpc`
        The MPC controller used as policy provider by this agent. The instance is
        modified in place to create the approximations of the state function
        :math:`V_\theta(s)` and action value function :math:`Q_\theta(s,a)`, so it is
        recommended not to modify it further after initialization of the agent.
        Moreover, some parameter and constraint names will need to be created, so an
        error is thrown if these names are already in use in the mpc.
    update_strategy : UpdateStrategy or int
        The strategy used to decide which frequency to update the mpc parameters with.
        If an ``int`` is passed, then the default strategy that updates every ``n``
        env's steps is used (where ``n`` is the argument passed); otherwise, an instance
        of :class:`core.update.UpdateStrategy` can be passed to specify the desired
        strategy in more details.
    discount_factor : float
        In RL, the factor that discounts future rewards in favor of immediate rewards.
        Usually denoted as :math:`\gamma`. It should satisfy :math:`\gamma \in (0, 1]`.
    optimizer : GradientBasedOptimizer
        A gradient-based optimizer (e.g., :class:`optim.GradientDescent`) to
        compute the updates of the learnable parameters, based on the current
        gradient-based RL algorithm.
    learnable_parameters : :class:`core.parameters.LearnableParametersDict`
        A special dict containing the learnable parameters of the MPC (usually referred
        to as :math:`\theta`), together with their bounds and values. This dict is
        complementary to :attr:`fixed_parameters`, which contains the MPC parameters
        that are not learnt by the agent.
    fixed_parameters : dict of (str, array_like) or collection of, optional
        A dict (or collection of dict, in case of the ``mpc`` wrapping an underlying
        :class:`csnlp.multistart.MultistartNlp` instance) whose keys are the names of
        the MPC parameters and the values are their corresponding values. Use this to
        specify fixed parameters, that is, non-learnable. If ``None``, then no fixed
        parameter is assumed.
    exploration : :class:`core.exploration.ExplorationStrategy`, optional
        Exploration strategy for inducing exploration in the online MPC policy. By
        default ``None``, in which case :class:`core.exploration.NoExploration` is used.
    experience : int or ExperienceReplay, optional
        The container for experience replay memory. If ``None`` is passed, then a memory
        with unitary length is created, i.e., it keeps only the latest memory
        transition. If an integer ``n`` is passed, then a memory with the length ``n``
        is created and with sample size ``n``. Otherwise, pass an instance of
        :class:`core.experience.ExperienceReplay` to specify the requirements in more
        details.
    warmstart : "last" or "last-successful" or WarmStartStrategy, optional
        The warmstart strategy for the MPC's NLP. If ``"last-successful"``, the last
        successful solution is used to warmstart the solver for the next iteration. If
        ``"last"``, the last solution is used, regardless of success or failure.
        Furthermore, an instance of :class:`core.warmstart.WarmStartStrategy` can
        be passed to specify a strategy for generating multiple warmstart points for the
        MPC's NLP instance. This is useful to generate multiple initial conditions for
        highly non-convex, nonlinear problems. This feature can only be used with an
        MPC that has an underlying multistart NLP problem (see :mod:`csnlp.multistart`).
    hessian_type : {"none", "approx", "full"}, optional
        The type of hessian to use in this (potentially) second-order algorithm.
        If ``"none"``, no second order information is used. If ``"approx"``, an easier
        approximation of it is used; otherwise, the full hessian is computed, but
        this is usually much more expensive. This option must be in accordance with the
        choice of ``optimizer``, that is, if the optimizer does not use second order
        information, then this option must be set to ``none``.
    record_td_errors: bool, optional
        If ``True``, the TD errors are recorded in the field :attr:`td_errors`, which
        otherwise is ``None``. By default, does not record them.
    use_last_action_on_fail : bool, optional
        In case the MPC solver fails
         - if ``False``, the action from the last solver's iteration is returned anyway
           (though suboptimal)
         - if ``True``, the action from the last successful call to the MPC is returned
           instead (if the MPC has been solved at least once successfully).

        By default, ``False``.
    remove_bounds_on_initial_action : bool, optional
        When ``True``, the upper and lower bounds on the initial action are removed in
        the action-value function approximator :math:`Q_\theta(s,a)` since the first
        action is constrained to be equal to the provided action :math:`a`. This is
        useful to avoid issues in the LICQ of the NLP. However, it can lead to numerical
        problems. By default, ``False``.
    name : str, optional
        Name of the agent. If ``None``, one is automatically created from a counter of
        the class' instancies.
    """

    def __init__(
        self,
        mpc: Mpc[SymType],
        update_strategy: Union[int, UpdateStrategy],
        discount_factor: float,
        optimizer: GradientBasedOptimizer,
        learnable_parameters: LearnableParametersDict[SymType],
        fixed_parameters: Union[
            None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]
        ] = None,
        exploration: Optional[ExplorationStrategy] = None,
        experience: Union[None, int, ExperienceReplay[ExpType]] = None,
        warmstart: Union[
            Literal["last", "last-successful"], WarmStartStrategy
        ] = "last-successful",
        hessian_type: Literal["none", "approx", "full"] = "approx",
        record_td_errors: bool = False,
        use_last_action_on_fail: bool = False,
        remove_bounds_on_initial_action: bool = False,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(
            mpc=mpc,
            update_strategy=update_strategy,
            discount_factor=discount_factor,
            learnable_parameters=learnable_parameters,
            optimizer=optimizer,
            fixed_parameters=fixed_parameters,
            exploration=exploration,
            experience=experience,
            warmstart=warmstart,
            use_last_action_on_fail=use_last_action_on_fail,
            remove_bounds_on_initial_action=remove_bounds_on_initial_action,
            name=name,
        )
        self.hessian_type = hessian_type
        self._sensitivity = self._init_sensitivity(hessian_type)
        self.td_errors: Optional[list[float]] = [] if record_td_errors else None

    def update(self) -> Optional[str]:
        sample = self.experience.sample()
        if self.hessian_type == "none":
            gradient = np.mean(list(sample), 0)
            return self.optimizer.update(gradient)

        gradients, hessians = zip(*sample)
        gradient = np.mean(gradients, 0)
        hessian = np.mean(hessians, 0)
        return self.optimizer.update(gradient, hessian)

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        truncated = terminated = False
        timestep = 0
        rewards = 0.0
        state = init_state
        action_space = getattr(env, "action_space", None)

        # solve for the first action
        action, solV = self.state_value(state, False, action_space=action_space)
        if not solV.success:
            self.on_mpc_failure(episode, None, solV.status, raises)

        while not (truncated or terminated):
            # compute Q(s,a)
            solQ = self.action_value(state, action)

            # step the system with action computed at the previous iteration
            new_state, cost, truncated, terminated, _ = env.step(action)
            self.on_env_step(env, episode, timestep)

            # compute V(s+) and store transition
            new_action, solV = self.state_value(
                new_state, False, action_space=action_space
            )
            if not self._try_store_experience(cost, solQ, solV):
                self.on_mpc_failure(
                    episode, timestep, f"{solQ.status} (Q); {solV.status} (V)", raises
                )

            # increase counters
            state = new_state
            action = new_action
            rewards += float(cost)
            timestep += 1
            self.on_timestep_end(env, episode, timestep)
        return rewards

    def train_one_rollout(
        self,
        rollout: Iterable[tuple[ObsType, ActType, float, ObsType]],
        episode: int,
        raises: bool = True,
    ) -> None:
        # in the case of off-policy q-learning, rollouts are made of
        # State-Action-Reward-next State (SARS) tuples
        for timestep, (state, action, cost, new_state) in enumerate(rollout, start=1):
            # compute Q(s,a)
            solQ = self.action_value(state, action)

            # compute V(s+) and store transition
            _, solV = self.state_value(new_state, False)
            if not self._try_store_experience(cost, solQ, solV):
                self.on_mpc_failure(
                    episode, timestep, f"{solQ.status} (Q); {solV.status} (V)", raises
                )
            self.on_timestep_end("off-policy", episode, timestep)

    def _init_sensitivity(
        self, hessian_type: Literal["none", "approx", "full"]
    ) -> Union[
        Callable[[cs.DM], np.ndarray], Callable[[cs.DM], tuple[np.ndarray, np.ndarray]]
    ]:
        """Internal utility to compute the derivative of ``Q(s,a)`` w.r.t. the learnable
        parameters, a.k.a., ``theta``."""
        ord = self.optimizer._order
        theta = cs.vvcat(self._learnable_pars.sym.values())
        nlp = self._Q.nlp
        x = nlp.x
        p = nlp.p
        lam_g_and_h = cs.vertcat(nlp.lam_g, nlp.lam_h)

        # compute first order sensitivity - necessary whatever the hessian type is
        snlp = NlpSensitivity(nlp, theta)
        gradient = snlp.jacobian("L-p")  # exact gradient, i.e., dQ/dtheta

        if hessian_type == "none":
            assert ord == 1, "Expected 1st-order optimizer with `hessian_type=none`."

            sensitivity = cs.Function(
                "lag_sens",
                [x, p, lam_g_and_h],
                [gradient],
                ["x", "p", "lam_g"],
                ["dQ"],
                {"cse": True},
            )

            def func(sol: Solution) -> np.ndarray:
                return np.asarray(sensitivity(sol.x, sol.p, sol.lam_g_and_h).elements())

        elif hessian_type == "approx":
            assert ord == 2, "Expected 2nd-order optimizer with `hessian_type=approx`."

            hessian = snlp.hessian("L-pp")  # approximate hessian

            # check if the hessian is not all zeros. If that's the case, we fall back to
            # computing just the gradient
            if hessian.nnz() > 0:
                sensitivity = cs.Function(
                    "lag_sens",
                    [x, p, lam_g_and_h],
                    [gradient, hessian],
                    ["x", "p", "lam_g"],
                    ["dQ", "ddQ"],
                    {"cse": True},
                )
                shape = sensitivity.size_out("ddQ")

                def func(sol: Solution) -> tuple[np.ndarray, np.ndarray]:
                    J, H = sensitivity(sol.x, sol.p, sol.lam_g_and_h)
                    return (
                        np.asarray(J.elements()),
                        np.reshape(H.elements(), shape, "F"),
                    )

            else:
                sensitivity = cs.Function(
                    "lag_sens",
                    [x, p, lam_g_and_h],
                    [gradient],
                    ["x", "p", "lam_g"],
                    ["dQ"],
                    {"cse": True},
                )

                def func(sol: Solution) -> tuple[np.ndarray, np.ndarray]:
                    J = sensitivity(sol.x, sol.p, sol.lam_g_and_h)
                    return np.asarray(J.elements()), 0.0

        else:
            assert ord == 2, "Expected 2nd-order optimizer with `hessian_type=full`."

            lam_lbx_and_ubx = cs.vertcat(nlp.lam_lbx, nlp.lam_ubx)
            Kp = snlp.jacobian("K-p")
            Ky = snlp.jacobian("K-y")
            dydtheta = -cs.solve(Ky, Kp)
            Lpy = cs.jacobian(gradient, nlp.primal_dual)
            hessian = snlp.hessian("L-pp") + Lpy @ dydtheta  # not sure if Lpy or Kp.T

            # check if the hessian is not all zeros. If that's the case, we fall back to
            # computing just the gradient
            if hessian.nnz() > 0:
                sensitivity = cs.Function(
                    "lag_sens",
                    [x, p, lam_g_and_h, lam_lbx_and_ubx],
                    [gradient, hessian],
                    ["x", "p", "lam_g", "lam_x"],
                    ["dQ", "ddQ"],
                    {"cse": True},
                )
                shape = sensitivity.size_out("ddQ")

                def func(sol: Solution) -> tuple[np.ndarray, np.ndarray]:
                    J, H = sensitivity(
                        sol.x, sol.p, sol.lam_g_and_h, sol.lam_lbx_and_ubx
                    )
                    return (
                        np.asarray(J.elements()),
                        np.reshape(H.elements(), shape, "F"),
                    )

            else:
                sensitivity = cs.Function(
                    "lag_sens",
                    [x, p, lam_g_and_h],  # should lam_lbx_and_ubx be included?
                    [gradient],
                    ["x", "p", "lam_g"],
                    ["dQ"],
                    {"cse": True},
                )

                def func(sol: Solution) -> tuple[np.ndarray, np.ndarray]:
                    J = sensitivity(sol.x, sol.p, sol.lam_g_and_h)
                    return np.asarray(J.elements()), 0.0

        return func

    def _try_store_experience(
        self, cost: SupportsFloat, solQ: Solution[SymType], solV: Solution[SymType]
    ) -> bool:
        """Internal utility that tries to store the gradient and hessian for the current
        transition in memory, if both ``V`` and ``Q`` were successful; otherwise, does
        not store it. Returns whether it was successful or not."""
        success = solQ.success and solV.success
        if success:
            td_error = cost + self.discount_factor * solV.f - solQ.f
            if self.hessian_type == "none":
                dQ = self._sensitivity(solQ)
                gradient = -td_error * dQ
                self.store_experience(gradient)
            else:
                dQ, ddQ = self._sensitivity(solQ)
                gradient = -td_error * dQ
                hessian = np.multiply.outer(dQ, dQ) - td_error * ddQ
                self.store_experience((gradient, hessian))
        else:
            td_error = np.nan

        if self.td_errors is not None:
            self.td_errors.append(td_error)
        return success



from typing import Any, Literal, Optional, TypeVar

import numpy as np
import numpy.typing as npt
from gymnasium import Env

from ...agents.common.learning_agent import ExpType, LearningAgent
from ...util.iters import bool_cycle
from ...util.seeding import RngType, mk_seed
from .wrapper import LearningWrapper, SymType

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


class Evaluate(LearningWrapper[SymType, ExpType]):
    """Wrapper for evaluating an agent during training.

    On the given hook and with the given frequency, this wrapper automatically evaluates
    the agent on the specified environment by calling the agent's method
    :meth:`mpcrl.Agent.evaluate`. The resulting evaluation returns are stored in the
    attribute :attr:`eval_returns`.

    Parameters
    ----------
    agent : LearningAgent
        The learning agent to be evaluated by the wrapper.
    eval_env : gymnasium.Env
        A gym environment to evaluate the agent in.
    hook : {"on_episode_end", "on_timestep_end", "on_update"}
        Hook to trigger the evaluation. The evaluation will be triggered every
        ``frequency`` invokations of the specified hook.
    frequency : int
        Frequency of the evaluation.
    n_eval_episodes : int, optional
        How many episodes to evaluate the agent for, by default ``1``.
    eval_immediately : bool, optional
        Whether to evaluate the agent immediately after the wrapper is created, by
        default ``False``.
    deterministic : bool, optional
        Whether the agent should act deterministically; by default, ``True``.
    seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
        Agent's and each env's random seeds for the evaluation.
    raises : bool, optional
        If ``True``, when any of the MPC solver runs fails, or when an update fails, the
        corresponding error is raised; otherwise, only a warning is raised.
    env_reset_options : dict, optional
        Additional information to specify how the environment is reset at each evalution
        episode (optional, depending on the specific environment).
    fix_seed : bool, optional
        If ``True``, the seed is fixed and the same seed is used for all evaluations.
    """

    def __init__(
        self,
        agent: LearningAgent[SymType, ExpType],
        eval_env: Env[ObsType, ActType],
        hook: Literal["on_episode_end", "on_timestep_end", "on_update"],
        frequency: int,
        n_eval_episodes: int = 1,
        eval_immediately: bool = False,
        *,
        deterministic: bool = True,
        seed: RngType = None,
        raises: bool = True,
        env_reset_options: Optional[dict[str, Any]] = None,
        fix_seed: bool = False,
    ) -> None:
        self.eval_env = eval_env
        self._hook = hook
        self._n_eval_episodes = n_eval_episodes
        self._deterministic = deterministic
        np_random = np.random.default_rng(seed)
        self._seed = mk_seed(np_random) if fix_seed else np_random
        self._raises = raises
        self._env_reset_options = env_reset_options
        self._keep_seed_fixed = fix_seed
        self._eval_cycle = bool_cycle(frequency)
        self.eval_returns: list[npt.NDArray[np.floating]] = []
        self._is_eval_in_progress = False
        super().__init__(agent)
        if eval_immediately:
            self._evaluate(force=True)

    def _evaluate(self, *_: Any, **kwargs: Any) -> None:
        unwrapped_agent = self.agent.unwrapped
        is_training = unwrapped_agent._is_training
        if (
            not is_training
            or self._is_eval_in_progress
            or (not kwargs.get("force", False) and not next(self._eval_cycle))
        ):
            return

        self._is_eval_in_progress = True
        try:
            self.eval_returns.append(
                self.agent.evaluate(
                    self.eval_env,
                    self._n_eval_episodes,
                    self._deterministic,
                    self._seed,
                    self._raises,
                    self._env_reset_options,
                )
            )
        finally:
            self._is_eval_in_progress = False
            unwrapped_agent._is_training = is_training

    def _establish_callback_hooks(self) -> None:
        super()._establish_callback_hooks()
        self._hook_callback(repr(self), self._hook, self._evaluate)



"""More often than not, the agent is given a nonlinear MPC problem to solve at each
time step. It is well-known that NLP problems are proned to convergence to local minima,
and a simple yet effective way to counter this is to solve the same problem multiple
times with different initial conditions. This is known as a warmstart strategy.

:class:`WarmStartStrategy` allows the user to specify a warmstart strategy for the NLP
solver of the MPC. In particular, the user can select which past solution to use as the
initial guess for the next iteration, and whether to generate structured or random
guesses around that solution to further explore the solution space and possibly converge
to a better solution. Moreover, the user can choose how this strategy should keep up
with new solutions.

This class makes heavy use of :mod:`csnlp.multistart`, so be sure to check the
documentation of that module for more basic information. See also
:ref:`user_guide_warmstarting`."""

from collections.abc import Generator
from itertools import chain
from typing import Literal, Optional

import numpy as np
from csnlp.multistart import RandomStartPoints, StructuredStartPoints
from numpy import typing as npt

from ..util.seeding import RngType


def _merge_init_conditions_dicts(
    init_cond: dict[str, npt.ArrayLike],
    prev_sol: dict[str, npt.ArrayLike],
) -> dict[str, npt.ArrayLike]:
    """Internal utility to merge two dictionaries of initial conditions, where the
    former comes from multistarting, and the latter from the previous solution."""
    out = prev_sol.copy()
    out.update(init_cond)
    return out


class WarmStartStrategy:
    """Class containing all the information to guide the warmstart strategy for the
    MPC's NLP in order to speed up computations (by selecting appropriate initial
    conditions) and to support multiple initial conditions.

    Parameters
    ----------
    warmstart: "last" or "last-successful", optional
        The warmstart strategy for the MPC's NLP solver. If ``"last-successful"``, the
        last **successful** solution is automatically used to warmstart the solver for
        the next iteration. If ``"last"``, the last solution is automatically used,
        regardless of success or failure.
    structured_points : :class:`csnlp.multistart.StructuredStartPoint`, optional
        Class containing info on how to generate structured starting points for the
        NLP solver. If ``None``, no structured point is generated.
    random_points : :class:`csnlp.multistart.RandomStartPoints`, optional
        Class containing info on how to generate random starting points for the NLP
        solver. Its :attr:`csnlp.multistart.RandomStartPoints.biases` field will be
        automatically updated with the last available solution (based on ``warmstart``),
        unless disabled with ``update_biases_for_random_points=False``). If ``None``,
        no random point is generated.
    update_biases_for_random_points : bool, optional
        If ``True``, the random points are biased around the values from the last
        available solution to the MPC optimization (based on ``warmstart``). If
        ``False``, the biases in ``random_points`` are not updated.
    seed : None, int, array of ints, SeedSequence, BitGenerator, Generator
        Seed for the random number generator. By default, ``None``.
    """

    def __init__(
        self,
        warmstart: Literal["last", "last-successful"] = "last-successful",
        structured_points: Optional[StructuredStartPoints] = None,
        random_points: Optional[RandomStartPoints] = None,
        update_biases_for_random_points: bool = True,
        seed: RngType = None,
    ) -> None:
        self.store_always = warmstart == "last"
        self.structured_points = structured_points
        self.random_points = random_points
        self.update_biases_for_random_points = update_biases_for_random_points
        self.reset(seed)

    @property
    def n_points(self) -> int:
        """Returns the number of both random and structured starting points."""
        return (0 if self.random_points is None else self.random_points.multistarts) + (
            0 if self.structured_points is None else self.structured_points.multistarts
        )

    def reset(self, seed: RngType = None) -> None:
        """Resets the :class:`numpy.random.Generator` for sampling random points, if any
        were specified."""
        if self.random_points is not None:
            self.random_points.np_random = np.random.default_rng(seed)

    def generate(
        self, previous_sol: Optional[dict[str, npt.ArrayLike]] = None
    ) -> Generator[dict[str, npt.ArrayLike], None, None]:
        """Generates some initial conditions/guesses for the primal optimization
        variables of the MPC's NLP problem.

        Parameters
        ----------
        previous_sol : dict of (str, array_like), optional
            Optional dict that contains the previous solution's values, if available. If
            passed, it is used

            - to update the random points' original, unless
              ``update_biases_for_random_points=False``, at which point the original
              biases are kept constant (this does not affect the generation of structure
              points in any way)
            - to fill in the rest of the initial conditions that are not included in the
              multistart strategy (neither structured nor random).

        Yields
        ------
        Generator of dict of (str, array_like)
            Yields the initial conditions for the MPC's NLP.
        """
        to_be_chained = []
        given_prev_sol = previous_sol is not None

        if self.structured_points is not None:
            to_be_chained.append(self.structured_points)

        if self.random_points is not None:
            if self.update_biases_for_random_points and given_prev_sol:
                self.random_points.biases.update(previous_sol)
            to_be_chained.append(self.random_points)

        generator = chain.from_iterable(to_be_chained)
        if given_prev_sol:
            generator = map(
                lambda ic: _merge_init_conditions_dicts(ic, previous_sol), generator
            )
        return generator

    def __repr__(self) -> str:
        nr = 0 if self.random_points is None else self.random_points.multistarts
        ns = 0 if self.structured_points is None else self.structured_points.multistarts
        return (
            f"{self.__class__.__name__}(store_always={self.store_always},"
            f"random_points={nr},structured_points={ns})"
        )

    def __str__(self) -> str:
        return self.__repr__()



"""A submodule with an utility class for assigning unique names to each instance of a
subclass. This is useful, for instance, for debugging and logging purposes when multiple
agents are trained in parallel."""

from collections.abc import Iterator as _Iterator
from itertools import count as _count
from typing import Optional


class Named:
    """Base class for objects with names. It assigns a unique name to each instance of a
    subclass by appending an ID (an incremental integer) to the class' name or the
    provided one.

    Parameters
    ----------
    name : str, optional
        Name of the object. If `None`, one is automatically created from a counter
        of the class' instancies.
    """

    __ids: dict[type, _Iterator[int]] = {}

    def __init__(self, name: Optional[str] = None) -> None:
        cls = self.__class__
        if cls in self.__ids:
            _id = self.__ids[cls]
        else:
            _id = _count(0)
            self.__ids[cls] = _id
        self.name = name or f"{cls.__name__}{next(_id)}"

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name}>"
